from abc import ABC, abstractmethod

from .keyboard import ButtonColors


class ItemContent(ABC):
    @abstractmethod
    def get_content(self, payload: str) -> dict:
        pass

    @abstractmethod
    def serialize(self) -> dict:
        pass


class TextItemContent(ItemContent):
    def __init__(self, text: str, color: int = ButtonColors.DEFAULT) -> None:
        self.text = text
        self.color = color

    def __repr__(self) -> str:
        return f'TextItemContent({repr(self.text)}, {self.color})'

    def get_content(self, payload: str) -> dict:
        return {
            'type': 'button',
            'text': self.text,
            'color': self.color,
            'payload': payload,
        }

    def serialize(self) -> dict:
        res = {
            'text': self.text,
        }

        if self.color != ButtonColors.DEFAULT:
            res['color'] = self.color

        return res



from typing import Any, Optional, Sequence


class ButtonColors:
    DEFAULT = 0
    PRIMARY = 1
    POSITIVE = 2
    NEGATIVE = 3


class KeyboardButton:
    def __init__(self, payload: Any, text: str, color: Optional[int] = ButtonColors.DEFAULT) -> None:
        self.payload = payload
        self.text = text
        self.color = color

    def __repr__(self) -> str:
        return f'KeyboardButton({repr(self.payload)}, {repr(self.text)}, {self.color})'


class Keyboard:
    def __init__(self, lines: Optional[Sequence] = None) -> None:
        if lines is None:
            lines = []

        self.lines = lines

    def __repr__(self) -> str:
        return f'Keyboard({self.lines})'

    def add_button(self, button: KeyboardButton) -> None:
        if len(self.lines) == 0:
            self.add_line()

        self.lines[-1].append(button)

    def add_line(self) -> None:
        self.lines.append([])



from abc import ABC, abstractmethod
from typing import Iterator, Optional

from .actions import Action
from .item_contents import ItemContent


class BaseItem(ABC):
    def __init__(self, name: str = None, action: Action = None) -> None:
        self.name = name
        self.action = action

    def __repr__(self) -> str:
        return f'BaseItem({repr(self.name)}, {self.action})'

    @abstractmethod
    def get_content(self) -> dict:
        pass

    def is_available(self, payload: Optional[dict] = None) -> bool:
        return True

    def on_select(self, payload: Optional[dict] = None) -> Iterator:
        if self.action is None:
            return ()

        actions = (self.action, ) if isinstance(self.action, Action) else self.action
        return (action.process(payload) for action in actions)

    def serialize(self) -> dict:
        res = {}

        if self.name is not None:
            res['name'] = self.name

        if self.action is not None:
            res['action'] = {
                'type': self.action.__class__.__name__,
                **self.action.serialize(),
            }

        return res


class Item(BaseItem):
    def __init__(self, name: str, content: ItemContent, action: Action = None) -> None:
        super().__init__(name, action)

        self.content = content

    def __repr__(self) -> str:
        return f'Item({repr(self.name)}, {self.content}, {self.action})'

    def get_content(self) -> dict:
        return self.content.get_content(self.name)

    def serialize(self) -> dict:
        res = super().serialize()

        res['content'] = {
            'type': self.content.__class__.__name__,
            **self.content.serialize(),
        }

        return res


class LineBreakItem(BaseItem):
    def __repr__(self) -> str:
        return 'LineBreakItem()'

    def get_content(self) -> dict:
        return {
            'type': 'line_break',
        }


class ConditionalItem(Item):
    def __init__(self, name: str, content: ItemContent, action: Action, condition: str) -> None:
        super().__init__(name, content, action)

        self.condition = condition

    def __repr__(self) -> str:
        return f'ConditionalItem({repr(self.name)}, {self.content}, {self.action}, {repr(self.condition)})'

    def is_available(self, payload: Optional[dict] = None) -> bool:
        return eval(self.condition)(payload)

    def serialize(self) -> dict:
        res = super().serialize()

        res['condition'] = self.condition

        return res



from abc import ABC, abstractmethod


class BaseContent(ABC):
    def __getitem__(self, key):
        return getattr(self, key)

    @abstractmethod
    def serialize(self) -> dict:
        pass


class Content(BaseContent):
    def __init__(self, text: str = None) -> None:
        self.text = text

    def __repr__(self) -> str:
        return f'Content({repr(self.text)})'

    @staticmethod
    def keys() -> tuple[str]:
        return 'text',

    def serialize(self) -> dict:
        res = {}

        if self.text is not None:
            res['text'] = self.text

        return res



from abc import ABC, abstractmethod
from typing import Optional


class StateHandler(ABC):
    @abstractmethod
    def get(self, user_id: Optional[int]) -> str:
        pass

    @abstractmethod
    def set(self, user_id: Optional[int], new_state: str) -> None:
        pass

    @abstractmethod
    def go_back(self, user_id: Optional[int], count: Optional[int] = 1) -> None:
        pass


class MemoryStateHandler(StateHandler):
    def __init__(self, default_state: str) -> None:
        self.default_state = default_state

        self.state = {}
        self.history = {}

    def __repr__(self) -> str:
        return f'MemoryStateHandler({self.default_state})'

    def get(self, user_id: Optional[int]) -> str:
        if user_id not in self.state:
            return self.default_state

        return self.state[user_id]

    def set(self, user_id: Optional[int], new_state: str) -> None:
        if user_id not in self.history:
            self.history[user_id] = []

        self.history[user_id].append(self.state[user_id] if user_id in self.state else self.default_state)
        self.state[user_id] = new_state

    def go_back(self, user_id: Optional[int], count: Optional[int] = 1) -> None:
        if count < 1:
            raise ValueError('Count must be at least 1')

        elif count == -1:
            count = 1000

        if user_id not in self.history:
            return

        for i in range(count):
            if len(self.history[user_id]) <= i + 1:
                new_state = self.default_state
                break

            else:
                new_state = self.history[user_id][-(i + 1)]

        self.state[user_id] = new_state
        self.history[user_id] = self.history[user_id][:-count]



from typing import Dict, Optional, Sequence

from .menus import BaseMenu
from .state import StateHandler
from .responses import Message, Response


class MenuManager:
    def __init__(self, menus: Dict[str, BaseMenu], state_handler: StateHandler) -> None:
        self.menus = menus
        self.state_handler = state_handler

    def __repr__(self) -> str:
        return f'MenuManager({self.menus}, {self.state_handler})'

    def get_message(self, user_id: int = None, payload: Optional[dict] = None) -> Message:
        state = self.state_handler.get(user_id)

        return self.menus[state].get_message(payload)

    def select(self, action: str, user_id: int = None, payload: Optional[dict] = None) -> Optional[Sequence[Message]]:
        state = self.state_handler.get(user_id)

        actions = self.menus[state].select(action, payload)
        if actions is not None:
            messages = []
            for res in actions:
                if isinstance(res, Message):
                    messages.append(res)

                elif isinstance(res, Response):
                    if res.message:
                        messages.append(res.message)

                    if res.go_back_count:
                        self.state_handler.go_back(user_id, res.go_back_count)

                    if res.menu:
                        self.state_handler.set(user_id, res.menu)

                        enter_res = self.menus[res.menu].enter(payload)
                        if isinstance(enter_res, Message):
                            messages.append(enter_res)

            return messages

        else:
            raise ValueError('Invalid action')

    def serialize(self) -> dict:
        return {
            'menus': {menu_name: {
                'type': menu.__class__.__name__,
                **menu.serialize(),
            } for menu_name, menu in self.menus.items()},
        }



import collections.abc
from abc import ABC, abstractmethod
from typing import Iterator, Optional, Sequence

from .actions import Action
from .contents import BaseContent
from .items import BaseItem
from .responses import Message
from .keyboard import KeyboardButton, Keyboard


class BaseMenu(ABC):
    @abstractmethod
    def select(self, action: str, payload=None) -> Optional[Iterator[Message]]:
        pass

    @abstractmethod
    def get_message(self, payload: Optional[dict] = None) -> Message:
        pass

    @abstractmethod
    def serialize(self) -> dict:
        pass

    def enter(self, payload: Optional[dict] = None) -> Optional[Message]:
        pass


class Menu(BaseMenu):
    def __init__(
            self, content: BaseContent, items: Sequence[BaseItem], default_action: Optional[Action] = None
    ) -> None:
        if items is None:
            items = []

        self.content = content
        self.items = items
        self.default_action = default_action

    def __repr__(self) -> str:
        return f'Menu({self.content}, {self.items}, {repr(self.default_action)})'

    def select(self, action: str, payload: Optional[dict] = None) -> Optional[Iterator[Message]]:
        target_item = next((
            i for i in self.items if (isinstance(i, BaseItem) and i.name == action and i.is_available(payload))
        ), None)

        if target_item is None:
            if self.default_action:
                return self.default_action.process(payload),

            return

        return target_item.on_select(payload)

    def get_message(self, payload: Optional[dict] = None) -> Message:
        if payload is None:
            payload = {}

        keyboard = Keyboard()
        for item in self.items:
            if item.is_available(payload):
                kwargs = item.get_content()

                if kwargs['type'] == 'button':
                    del kwargs['type']

                    if 'text' in kwargs:
                        kwargs['text'] = kwargs['text'].format(**payload)

                    keyboard.add_button(KeyboardButton(**kwargs))

                elif kwargs['type'] == 'line_break':
                    keyboard.add_line()

        return Message(payload=payload, keyboard=keyboard, **self.content)

    def serialize(self) -> dict:
        res = {
            'content': {
                'type': self.content.__class__.__name__,
                **self.content.serialize(),
            }
        }

        if self.items:
            res['items'] = [{
                'type': item.__class__.__name__,
                **item.serialize(),
            } for item in self.items]

        if self.default_action:
            res['default_action'] = {
                'type': self.default_action.__class__.__name__,
                **self.default_action.serialize(),
            }

        return res


class CustomMenu(BaseMenu):
    def __init__(self, handler) -> None:
        super().__init__()

        self.handler = handler

    def __repr__(self) -> str:
        return f'CustomMenu({self.handler})'

    def select(self, action: str, payload: Optional[dict] = None) -> Optional[Iterator[Message]]:
        res = self.handler.select(action, payload)

        return res if isinstance(res, collections.abc.Sequence) else (res, )

    def get_message(self, payload: Optional[dict] = None) -> Optional[Message]:
        return self.handler.get_message(payload)

    def serialize(self) -> dict:
        res = {
            'handler': self.handler.__name__
        }

        return res

    def enter(self, payload: Optional[dict] = None) -> Optional[Message]:
        return self.handler.enter(payload)



from typing import Optional

from navmenu.keyboard import Keyboard


class Message:
    def __init__(
            self, text: Optional[str] = None, keyboard: Optional[Keyboard] = None, payload: Optional[dict] = None
    ) -> None:
        if payload is None:
            payload = {}

        self.keyboard = keyboard

        self.text = text.format(**payload) if text is not None else None

    def __repr__(self) -> str:
        return f'Message({repr(self.text)}, {self.keyboard})'


class Response:
    def __init__(self, message: Optional[Message] = None, menu: str = None, go_back_count: int = None) -> None:
        self.message = message
        self.menu = menu
        self.go_back_count = go_back_count

    def __repr__(self) -> str:
        return f'Response({self.message}, {repr(self.menu)}, {self.go_back_count})'



from types import ModuleType
from typing import Optional, Sequence

from . import actions
from . import contents
from . import item_contents
from . import items
from . import menus


def filter_kwargs(data: dict, extra_excludes: Optional[Sequence] = None) -> dict:
    if extra_excludes is None:
        extra_excludes = ()

    return {k: v for k, v in data.items() if k not in ['type', *extra_excludes]}


def deserialize_action(data: dict, function_container: ModuleType):
    class_ = getattr(actions, data['type'])

    if 'function' in data:
        return class_(
            **filter_kwargs(data, ('function', )),
            function=getattr(function_container, data['function']),
        )

    else:
        return class_(**filter_kwargs(data))


def deserialize_item_content(data: dict):
    return getattr(item_contents, data['type'])(**filter_kwargs(data))


def deserialize_item(data: dict, function_container: ModuleType):
    kwargs = filter_kwargs(data, ('action', 'content'))

    if 'action' in data:
        kwargs['action'] = deserialize_action(data['action'], function_container)

    if 'content' in data:
        kwargs['content'] = deserialize_item_content(data['content'])

    return getattr(items, data['type'])(**kwargs)


def deserialize_content(data: dict):
    return getattr(contents, data['type'])(**filter_kwargs(data))


def deserialize_menu(data: dict, function_container: ModuleType, custom_menu_handlers):
    class_ = getattr(menus, data['type'])

    if 'items' in data:
        return class_(
            **filter_kwargs(data, ('content', 'items', 'default_action')),
            content=deserialize_content(data['content']),
            items=[deserialize_item(i, function_container) for i in data['items']],
            default_action=(
                deserialize_action(data['default_action'], function_container) if 'default_action' in data else None
            ),
        )

    else:
        return class_(
            **filter_kwargs(data, ('handler', )),
            handler=next(i for i in custom_menu_handlers if i.__name__ == data['handler']),
        )


def deserialize(data: dict, function_container: ModuleType = None, custom_menu_handlers: Sequence = None) -> dict:
    return {
        menu_name: deserialize_menu(menu, function_container, custom_menu_handlers)
        for menu_name, menu in data['menus'].items()
    }



from abc import ABC, abstractmethod
from typing import Optional, Union

from .responses import Message, Response


class Action(ABC):
    @abstractmethod
    def process(self, payload: Optional[dict] = None) -> Message:
        pass

    @abstractmethod
    def serialize(self) -> dict:
        pass


class MessageAction(Action):
    def __init__(self, text: str) -> None:
        self.text = text

    def __repr__(self) -> str:
        return f'MessageAction({repr(self.text)})'

    def process(self, payload: Optional[dict] = None) -> Message:
        return Message(self.text, payload=payload)

    def serialize(self) -> dict:
        return {
            'text': self.text,
        }


class SubmenuAction(Action):
    def __init__(self, menu_name: str) -> None:
        self.menu_name = menu_name

    def __repr__(self) -> str:
        return f'SubmenuAction({repr(self.menu_name)})'

    def process(self, payload: Optional[dict] = None) -> Response:
        return Response(menu=self.menu_name)

    def serialize(self) -> dict:
        return {
            'menu_name': self.menu_name,
        }


class GoBackAction(Action):
    def __init__(self, count: int = 1) -> None:
        self.count = count

    def __repr__(self) -> str:
        return f'GoBackAction({self.count})'

    def process(self, payload: Optional[dict] = None) -> Response:
        return Response(go_back_count=self.count)

    def serialize(self) -> dict:
        res = {}

        if self.count != 1:
            res['count'] = self.count

        return res


class ExecuteAction(Action):
    def __init__(self, command: str, return_text: bool = False) -> None:
        self.command = command
        self.return_text = return_text

    def __repr__(self) -> str:
        return f'ExecuteAction({repr(self.command)}, {self.return_text})'

    def process(self, payload: Optional[dict] = None) -> Message:
        return Message(str(eval(self.command)) if self.return_text else exec(self.command), payload=payload)

    def serialize(self) -> dict:
        res = {
            'command': self.command,
        }

        if self.return_text is not False:
            res['return_text'] = self.return_text

        return res


class FunctionAction(Action):
    def __init__(self, function) -> None:
        self.function = function

    def __repr__(self) -> str:
        return f'FunctionAction({self.function})'

    def process(self, payload: Optional[dict] = None) -> Union[Message, Response]:
        return self.function(payload)

    def serialize(self) -> dict:
        return {
            'function': self.function.__name__,
        }



from setuptools import find_packages, setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='navmenu',
    version='0.1.0',
    description='A library to create multilevel menus for chatbots',
    long_description=long_description,
    author='rashidsh',
    author_email='mail@rashidsh.ru',
    url='https://github.com/rashidsh/navmenu',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(),
    python_requires='>=3.6',
    tests_require=[
        'pytest',
    ],
)
